/*
 * log.c - message logging functions
 *
 * Grigoriy Sitkarev, <sitkarev@unixkomi.ru>
 */

/*
 * 12.09.04 created first version
 */

#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>

#include "macros.h"
#include "log.h"

char *__program_name;		/* running program name */

static int log_facility = 0;	/* syslog facility used */
static int log_to_syslog = 0;	/* enable or disable logging to syslog */
static log_level_t log_level = LOG_LEVEL_DEBUG;
static char *time_format = NULL; /* strftime timestamp format */

static error_handler_t error_handler = NULL;

void set_error_hanlder(error_handler_t handler)
{
	if (handler != NULL)
		error_handler = handler;
	else
		error_handler = NULL;
}

void log_init(char *argv0, log_level_t level, int to_syslog, int facility, char *opt_time_format)
{
	char *str;

	if (argv0 == NULL) {
		fprintf(stderr, "Program name not available in argv0!\n");
		exit(-1);
	} else {
		__program_name = (str = strrchr(argv0, '/')) ? ++str : argv0;
	}

	switch (level) {
	case LOG_LEVEL_DEBUG:
	case LOG_LEVEL_VERBOSE:
	case LOG_LEVEL_ERROR:
	case LOG_LEVEL_WARN:
	case LOG_LEVEL_INFO:
		log_level = level;
		break;
	default:
		fprintf(stderr, "Unknown log level %d\n", level);
		exit(-1);
		break;
	}

	if (to_syslog) {

		log_to_syslog = 1;

		switch (facility) {
		case LOG_FACILITY_DAEMON:
			log_facility = LOG_DAEMON;
			break;
		case LOG_FACILITY_LOCAL0:
			log_facility = LOG_LOCAL0;
			break;
		case LOG_FACILITY_LOCAL1:
			log_facility = LOG_LOCAL1;
			break;
		case LOG_FACILITY_LOCAL2:
			log_facility = LOG_LOCAL2;
			break;
		case LOG_FACILITY_LOCAL3:
			log_facility = LOG_LOCAL3;
			break;
		case LOG_FACILITY_LOCAL4:
			log_facility = LOG_LOCAL4;
			break;
		case LOG_FACILITY_LOCAL5:
			log_facility = LOG_LOCAL5;
			break;
		case LOG_FACILITY_LOCAL6:
			log_facility = LOG_LOCAL6;
			break;
		case LOG_FACILITY_LOCAL7:
			log_facility = LOG_LOCAL7;
			break;
		default:
			fprintf(stderr, "Unknown syslog facility code %d\n",
					facility);
			exit(-1);
		}
	}

	if (opt_time_format != NULL) {

		if (time_format)
			free(time_format);

		time_format = strdup(opt_time_format);

		if (!time_format) {
			fprintf(stderr, "No memory to store time format\n");
			exit(-1);
		}
	}
}

static void do_log(log_level_t level, const char *fmt, va_list list)
{
	time_t secs;
	struct tm *tm;
	char *txt;
	char stamp_buffer[64];
	char prefix_buffer[64];
	char fmt_buffer[1024];
	char msg_buffer[1024];
	int saved_errno;
	int prio;
	va_list va_save;

	saved_errno = errno;

	if (level != LOG_LEVEL_ERROR && level > log_level)
		return;

	switch (level) {
	case LOG_LEVEL_DEBUG:
		txt = "debug";
		prio = LOG_DEBUG;
		break;
	case LOG_LEVEL_VERBOSE:
		txt = "verbose";
		prio = LOG_INFO;
		break;
	case LOG_LEVEL_INFO:
		txt = "info";
		prio = LOG_INFO;
		break;
	case LOG_LEVEL_WARN:
		txt = "warn";
		prio = LOG_WARNING;
		break;
	case LOG_LEVEL_ERROR:
		txt = "error";
		prio = LOG_ERR;
		break;
	default:
		txt = "unknown";
		prio = LOG_ERR;
		break;
	}

	if (__program_name != NULL)
		snprintf(prefix_buffer, sizeof(prefix_buffer),
				"%s <%s>", __program_name, txt);
	else
		snprintf(prefix_buffer, sizeof(prefix_buffer), "<%s>", txt);

	if (time_format) {
		secs = time(NULL);
		tm = localtime(&secs);
		strftime(stamp_buffer, sizeof(stamp_buffer), time_format, tm);

		snprintf(fmt_buffer, sizeof(fmt_buffer), "%s %s %s",
				stamp_buffer, prefix_buffer, fmt);
	} else
		snprintf(fmt_buffer, sizeof(fmt_buffer), "%s %s",
				prefix_buffer, fmt);

	va_copy(va_save, list);
	vsnprintf(msg_buffer, sizeof(msg_buffer), fmt_buffer, list);

	if (log_to_syslog) {
		openlog(__program_name ? __program_name : NULL,
				LOG_NDELAY | LOG_PID, log_facility);
		vsyslog(prio, fmt, va_save);
		closelog();
	} else {
		snprintf(fmt_buffer, sizeof(fmt_buffer), "%s\r\n", msg_buffer);
		write(STDERR_FILENO, fmt_buffer, strlen(fmt_buffer));
	}
	va_end(va_save);
	errno = saved_errno;
}

void debug(const char *fmt, ...)
{
	va_list list;

	va_start(list, fmt);
	do_log(LOG_LEVEL_DEBUG, fmt, list);
	va_end(list);
}

void verbose(const char *fmt, ...)
{
	va_list list;

	va_start(list, fmt);
	do_log(LOG_LEVEL_VERBOSE, fmt, list);
	va_end(list);
}

void warn(const char *fmt, ...)
{
	va_list list;

	va_start(list, fmt);
	do_log(LOG_LEVEL_WARN, fmt, list);
	va_end(list);
}

void info(const char *fmt, ...)
{
	va_list list;

	va_start(list, fmt);
	do_log(LOG_LEVEL_INFO, fmt, list);
	va_end(list);
}

void error(const char *fmt, ...)
{
	va_list list;

	va_start(list, fmt);
	do_log(LOG_LEVEL_ERROR, fmt, list);
	va_end(list);

	if (error_handler != NULL) {
		(*error_handler)(-1);
	} else {
		exit(-1);
	}
}

void error_exit(int code, const char *fmt, ...)
{
	va_list list;

	va_start(list, fmt);
	do_log(LOG_LEVEL_ERROR, fmt, list);
	va_end(list);

	if (error_handler != NULL) {
		(*error_handler)(code);
	} else {
		exit(code);
	}
}
